/***************************************************************************
 *   $Id: htmltpl.c,v 1.1.1.1 2011-11-17 16:42:28 scg Exp $
 *                                                                         *
 *   (C)2011 St(u)dio of Computer Games                                    *
 *   Alexander Ozumenko <a.ozumenko@gmail.com>                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <assert.h>

#include "variable.h"
#include "parse.h"
#include "list.h"



#define CMD_PAGE        512
#define STACK_PAGE      64

#define ALIGN(v, b)     ((((v) + (b) - 1) / (b)) * (b))



/* template block structure */
struct tpl_cmd {
    size_t size;
#define CMD_TEXT        0
#define CMD_VAR         1
#define CMD_JMP_IF      2
#define CMD_JMP_UNLESS  3
#define CMD_JMP         4
#define CMD_LOOP_INIT   5
#define CMD_LOOP        6
    int type;

    union {
        struct cmd_text {
            char data[0];
        } text;
        struct cmd_var {
            char name[0];
        } var;
        struct cmd_jmp_if {
            int addr;
            char name[0];
        } jmp_if;
        struct cmd_jmp {
            int addr;
        } jmp;
        struct cmd_loop_init {
            char name[0];
        } loop_init;
        struct cmd_loop {
            int addr;
        } loop;
    };
};


/* template stack element */
enum STACK_TYPE {
    STACK_C_IF,
    STACK_C_UNLESS,
    STACK_C_LOOP,
    STACK_I_LOOP,
};

struct tpl_stack {
    enum STACK_TYPE type;

    union {
        /* compilator */
        struct stack_c_if {
            int addr_if, addr_else;
        } c_if;
        struct stack_c_loop {
            int addr;
        } c_loop;

        /* interpretator */
        struct stack_i_loop {
            struct variable *var_current;
            struct list_head *head, *pos;
        } i_loop;
    };
};


/* template structure */
struct tpl {
    /* template commands block */
    char *cmd;
    int cmd_size, cmd_length;

    /* stack */
    struct tpl_stack *stack;
    int stack_size, stack_length;

    /* variables enviroment */
    struct enviroment env;

    /* interpretator state */
    struct variable *var_current;

    /* last error state */
    struct error {
        int n;
        enum parse_tag tag, stack_tag;
        int col, row;
        char *filename;
    } error;
};


/* FixMe: ... */
static const char *tag_name[] = {
    [TAG_VAR] = "<TMPL_VAR>",
    [TAG_LOOP] = "<TMPL_LOOP>",
    [TAG_LOOP_CLOSE] = "</TMPL_LOOP>",
    [TAG_INCLUDE] = "<TMPL_INCLUDE>",
    [TAG_IF] = "<TMPL_IF>",
    [TAG_IF_CLOSE] = "</TMPL_IF>",
    [TAG_UNLESS] = "<TMPL_UNLESS>",
    [TAG_UNLESS_CLOSE] = "</TMPL_UNLESS>",
    [TAG_ELSE] = "<TMPL_ELSE>"
};

static const int stack_to_tag[] = {
    [STACK_C_IF] = TAG_IF_CLOSE,
    [STACK_C_UNLESS] = TAG_UNLESS_CLOSE,
    [STACK_C_LOOP] = TAG_LOOP_CLOSE,
    [STACK_I_LOOP] = TAG_LOOP_CLOSE // FixMe: что-то не то
};


/*
 * Stack section
 */

static inline void tpl_stack_reset(struct tpl *tpl) {
    tpl->stack_length = 0;
}


static inline struct tpl_stack *tpl_stack_push(struct tpl *tpl) {
    if(tpl->stack_length == tpl->stack_size) {
        int size;
        void *data;

        size = tpl->stack_size + STACK_PAGE;
        data = realloc(tpl->stack, size * sizeof(struct tpl_stack));
        if(data == NULL) {
            return NULL;
        }

        tpl->stack = data;
        tpl->stack_size = size;
    }

    return &tpl->stack[tpl->stack_length++];
}


static inline struct tpl_stack *tpl_stack_last(struct tpl *tpl) {
    return &tpl->stack[tpl->stack_length - 1];
}


static inline struct tpl_stack *tpl_stack_pop(struct tpl *tpl) {
    return tpl->stack_length > 0 ? &tpl->stack[--tpl->stack_length] : NULL;
}



/*
 * Template commands section
 */

/* get next free template command */
static inline struct tpl_cmd *tpl_cmd_add(struct tpl *tpl, int type, size_t data_size) {
    size_t cmd_size;
    struct tpl_cmd *cmd;

    /* calculate commanbd size */
    cmd_size = sizeof(struct tpl_cmd) + data_size;

    /* align command size to int length */
    cmd_size = ALIGN(cmd_size, sizeof(int));

    /* reallocation pattern data buffer */
    if(tpl->cmd_length + cmd_size >= tpl->cmd_size) {
        int size;
        void *data;

        size = ALIGN(tpl->cmd_size + cmd_size, CMD_PAGE);

        data = realloc(tpl->cmd, size);
        if(data == NULL) {
            return NULL;
        }

        tpl->cmd = data;
        tpl->cmd_size = size;
    }

    cmd = (struct tpl_cmd *)(tpl->cmd + tpl->cmd_length);
    tpl->cmd_length += cmd_size;

    cmd->size = cmd_size;
    cmd->type = type;

    return cmd;
}


/* get cmd address */
static inline int tpl_cmd_addr(struct tpl *tpl, struct tpl_cmd *cmd) {
    return (char *)cmd - tpl->cmd;
}


/* get cmd by address */
static inline struct tpl_cmd *tpl_cmd_get(struct tpl *tpl, int addr) {
    return (struct tpl_cmd *)(tpl->cmd + addr);
}



/*
 * Parsing section
 */

/* tag hooks declarations */
static int tpl_tag_hook(struct parse *parse, enum parse_tag tag, const char *var);
static int tpl_data_hook(struct parse *parse, const char *data, size_t len);


/* FixME: ... */
static void tpl_make_filename_error(struct parse *parse, const char *filename) {
    struct tpl *tpl;

    tpl = (struct tpl *)parse->user_data;

    if(tpl->error.filename == NULL) {
        tpl->error.filename = strdup(filename);
    }
}


/* make a parser syntax error */
static void tpl_make_syntax_error(struct parse *parse, enum parse_tag tag, struct tpl_stack *stack) {
    struct tpl *tpl;

    tpl = (struct tpl *)parse->user_data;

    tpl->error.n = -1;    /* syntax error */
    tpl->error.tag = tag;
    tpl->error.stack_tag = stack == NULL ? TAG_NONE : stack_to_tag[stack->type];
    tpl->error.col = parse_tag_col(parse);
    tpl->error.row = parse_tag_row(parse);
}


/* load pattern file to exists template structure */
static int tpl_load_file(struct tpl *tpl, const char *filename) {
    struct parse parse;
    int fd;
    char buf[256];
    int n, err;

    parse_init(&parse, tpl, tpl_tag_hook, tpl_data_hook);

    err = parse_make_work_dir(&parse, filename);
    if(err) {
        return err;
    }

    fd = open(filename, O_RDONLY);
    if(fd < 0) {
        tpl_make_filename_error(&parse, filename);
        parse_release(&parse);
        return -1;
    }

    while((n = read(fd, buf, sizeof(buf))) > 0) {
        err = parse_routine(&parse, buf, n);
        if(err) {
            tpl_make_filename_error(&parse, filename);
            parse_release(&parse);
            close(fd);
            return err;
        }
    }

    close(fd);

    err = parse_release(&parse);
    if(err) {
        tpl_make_filename_error(&parse, filename);
    }

    return err;
}


/* parser tag hook */
static int tpl_tag_hook(struct parse *parse, enum parse_tag tag, const char *var) {
    struct tpl *tpl;
    struct tpl_cmd *cmd;
    struct tpl_stack *stack;

    tpl = (struct tpl *)parse->user_data;

    switch(tag) {
    case TAG_VAR:
        cmd = tpl_cmd_add(tpl, CMD_VAR, strlen(var) + 1);
        if(cmd == NULL) {
            return -1;
        }
        strcpy(cmd->var.name, var);

        break;

    case TAG_LOOP:
        cmd = tpl_cmd_add(tpl, CMD_LOOP_INIT, strlen(var) + 1);
        if(cmd == NULL) {
            return -1;
        }
        strcpy(cmd->loop_init.name, var);

        cmd = tpl_cmd_add(tpl, CMD_LOOP, 0);
        if(cmd == NULL) {
            return -1;
        }

        stack = tpl_stack_push(tpl);
        if(stack == NULL) {
            return -1;
        }
        stack->type = STACK_C_LOOP;
        stack->c_loop.addr = tpl_cmd_addr(tpl, cmd);

        break;

    case TAG_LOOP_CLOSE:
        stack = tpl_stack_pop(tpl);
        if(stack == NULL || stack->type != STACK_C_LOOP) {
            tpl_make_syntax_error(parse, tag, stack);
            return -1;
        }

        cmd = tpl_cmd_add(tpl, CMD_JMP, 0);
        if(cmd == NULL) {
            return -1;
        }
        cmd->jmp.addr = stack->c_loop.addr;

        cmd = tpl_cmd_get(tpl, stack->c_loop.addr);
        cmd->loop.addr = tpl->cmd_length;

        break;

    case TAG_INCLUDE:
        {
            int err;
            const char *filename;
            char filename_buf[PATH_MAX + 1];
            if(var[0] != '/') {
                snprintf(filename_buf, sizeof(filename_buf), "%s/%s", parse->work_dir, var);
                filename = filename_buf;
            }
            else {
                filename = var;
            }
            err = tpl_load_file(tpl, filename);

            if(err) {
                // FixMe: как-то нужно отметить ошибку загрузки файла
fprintf(stderr, "не удалось загрузить файл <TMP_INCLUDE> %s [%d]: %d:%d\n",
    filename, err, parse_tag_row(parse), parse_tag_col(parse));
                return err;
            }
        }
        break;

    case TAG_IF:
    case TAG_UNLESS:
        stack = tpl_stack_push(tpl);
        if(stack == NULL) {
            return -1;
        }
        stack->c_if.addr_if = tpl->cmd_length;
        stack->c_if.addr_else = -1;

        if(tag == TAG_IF) {
            stack->type = STACK_C_IF;
            cmd = tpl_cmd_add(tpl, CMD_JMP_IF, strlen(var) + 1);
        }
        else {
            stack->type = STACK_C_UNLESS;
            cmd = tpl_cmd_add(tpl, CMD_JMP_UNLESS, strlen(var) + 1);
        }
        if(cmd == NULL) {
            return -1;
        }
        strcpy(cmd->jmp_if.name, var);

        break;

    case TAG_IF_CLOSE:
    case TAG_UNLESS_CLOSE:
        stack = tpl_stack_pop(tpl);
        if(tag == TAG_UNLESS_CLOSE && (stack == NULL || stack->type != STACK_C_UNLESS)) {
            tpl_make_syntax_error(parse, tag, stack);
            return -1;
        }
        else if(tag == TAG_IF_CLOSE && (stack == NULL || stack->type != STACK_C_IF)) {
            tpl_make_syntax_error(parse, tag, stack);
            return -1;
        }

        cmd = tpl_cmd_get(tpl, stack->c_if.addr_else < 0 ? stack->c_if.addr_if : stack->c_if.addr_else);
        cmd->jmp_if.addr = tpl->cmd_length;

        break;

    case TAG_ELSE:
        cmd = tpl_cmd_add(tpl, CMD_JMP, 0);
        if(cmd == NULL) {
            return -1;
        }

        stack = tpl_stack_last(tpl);
        if(stack == NULL || (stack->type != STACK_C_IF && stack->type != STACK_C_UNLESS)) {
            tpl_make_syntax_error(parse, tag, stack);
            return -1;
        }
        stack->c_if.addr_else = tpl_cmd_addr(tpl, cmd);

        cmd = tpl_cmd_get(tpl, stack->c_if.addr_if);
        cmd->jmp_if.addr = tpl->cmd_length;

        break;
    }

    return 0;
}


/* parser data hook */
static int tpl_data_hook(struct parse *parse, const char *data, size_t len) {
    struct tpl_cmd *cmd;

    cmd = tpl_cmd_add((struct tpl *)parse->user_data, CMD_TEXT, len + 1);
    if(cmd == NULL) {
        return -1;
    }

    memcpy(cmd->text.data, data, len);
    cmd->text.data[len] = '\0';

    return 0;
}



/*
 * Interfase section
 */

/* create template from file */
struct tpl *tpl_create(void) {
    struct tpl *tpl;
    int err;

    tpl = malloc(sizeof(struct tpl));
    if(tpl == NULL) {
        return NULL;
    }
    memset(tpl, 0, sizeof(struct tpl));

    tpl->cmd = NULL;
    tpl->cmd_length = tpl->cmd_size = 0;

    err = enviroment_init(&tpl->env);
    if(err) {
        free(tpl);
        return NULL;
    }

    /* init stack */
    tpl->stack = NULL;
    tpl->stack_length = tpl->stack_size = 0;

    /* interpretator init */
    tpl->var_current = tpl->env.root;

    /* init error state */
    tpl->error.n = 0;
    tpl->error.tag = tpl->error.stack_tag = 0;
    tpl->error.col = tpl->error.row = 0;
    tpl->error.filename = NULL;

    return tpl;
}


/* load and parse template from file */
int tpl_parse_file(struct tpl *tpl, const char *filename) {
    int err;

    err = tpl_load_file(tpl, filename);
    if(err) {
        return err;
    }

    tpl_stack_reset(tpl);

    return 0;
}


/* release template data */
void tpl_release(struct tpl *tpl) {
    if(tpl->error.filename != NULL) {
        free(tpl->error.filename);
    }

    if(tpl->cmd != NULL) {
        free(tpl->cmd);
    }

    if(tpl->stack != NULL) {
        free(tpl->stack);
    }

    enviroment_release(&tpl->env);

    free(tpl);
}


/* output template */
int tpl_output(struct tpl *tpl, FILE *stream) {
    struct tpl_cmd *cmd;
    struct variable *var;
    const char *value;
    struct tpl_stack *stack;
    int addr;

    if(tpl->cmd == NULL) {
        return 0;
    }

    addr = 0;
    while(addr < tpl->cmd_length) {
        cmd = tpl_cmd_get(tpl, addr);

        switch(cmd->type) {
        case CMD_TEXT:
            fprintf(stream, "%s", cmd->text.data);
            addr += cmd->size;
            break;

        case CMD_VAR:
            value = variable_scalar_get(variable_hash_get(tpl->var_current, cmd->var.name));
            if(value != NULL) {
                fprintf(stream, "%s", value);
            }
            addr += cmd->size;
            break;

        case CMD_JMP_IF:
            var = variable_hash_get(tpl->var_current, cmd->jmp_if.name);
            if(var == NULL) {
                addr = cmd->jmp_if.addr;
            }
            else {
                addr += cmd->size;
            }
            break;

        case CMD_JMP_UNLESS:
            var = variable_hash_get(tpl->var_current, cmd->jmp_if.name);
            if(var == NULL) {
                addr += cmd->size;
            }
            else {
                addr = cmd->jmp_if.addr;
            }
            break;

        case CMD_JMP:
            addr = cmd->jmp.addr;
            break;

        case CMD_LOOP_INIT:
            stack = tpl_stack_push(tpl);
            if(stack == NULL) {
                return -1;
            }
            stack->type = STACK_I_LOOP;

            var = variable_hash_get(tpl->var_current, cmd->loop_init.name);
            if(var != NULL && variable_type(var) == VARIABLE_TYPE_ARRAY) {
                stack->i_loop.var_current = tpl->var_current;
                stack->i_loop.head = &var->array.list;
                stack->i_loop.pos = stack->i_loop.head->next;
            }
            else {
                stack->i_loop.head = NULL;
            }
            addr += cmd->size;
            break;

        case CMD_LOOP:
            stack = tpl_stack_last(tpl);
            assert(stack != NULL);
            if(stack->i_loop.head != NULL && stack->i_loop.pos != stack->i_loop.head) {
                tpl->var_current = list_entry(stack->i_loop.pos, struct variable, placement.list);
                stack->i_loop.pos = stack->i_loop.pos->next;
                addr += cmd->size;
            }
            else {
                addr = cmd->loop.addr;
                tpl->var_current = stack->i_loop.var_current;
                tpl_stack_pop(tpl);
            }
            break;
        }
    }

    return 0;
}


/* get a last error */
const char *tpl_error(struct tpl *tpl) {
    static char buf[256];
    const char *filename;

    filename = tpl->error.filename == NULL ? "no file" : tpl->error.filename;

    if(tpl->error.n == -1) {
        snprintf(buf, sizeof(buf), "met %s, when %s was expected at %s:%d:%d",
            tag_name[tpl->error.tag],
            tpl->error.stack_tag == TAG_NONE ? "nothing" : tag_name[tpl->error.stack_tag],
            filename,
            tpl->error.row, tpl->error.col);
    }
    else {
        snprintf(buf, sizeof(buf), "%s: %s", strerror(errno), filename);
    }

    return buf;
}



/**
 * Variables interface section
 */

/* get template enviroment */
struct enviroment *tpl_get_enviroment(struct tpl *tpl) {
    return &tpl->env;
}
