/***************************************************************************
 *   $Id: parse.c,v 1.1.1.1 2011-11-17 16:42:28 scg Exp $
 *                                                                         *
 *   (C)2009 St(u)dio of Computer Games                                    *
 *   Alexander Ozumenko <a.ozumenko@gmail.com>                             *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <string.h>
#include <stdlib.h>
#include <libgen.h>
#include <errno.h>

#include "parse.h"



#define NAME_BLOCK_PAGE         512
#define IS_WHITESPACE(x)        ' ' == (x) || '\t' == (x) || '\r' == (x) || '\n' == (x)
#define NOT_WHITESPACE(x)       ' ' != (x) && '\t' != (x) && '\r' != (x) && '\n' != (x)


#define BUF_IS_NOT_EMPTY(parse) \
    parse->buf_pos > 0

#define BUF_RESET(parse) \
    parse->buf_pos = 0; \
    parse->name_start = -1

#define BUFF_APPEND(parse, data, length) \
do { \
    if(parse->buf_pos + length >= parse->buf_size) { \
        parse->buf_size = (((parse->buf_size + length) / NAME_BLOCK_PAGE) + 1) * NAME_BLOCK_PAGE; \
        parse->buf = realloc(parse->buf, parse->buf_size); \
    } \
    strncpy(parse->buf + parse->buf_pos, data, length); \
    parse->buf_pos += length; \
    parse->buf[parse->buf_pos] = '\0'; \
} while(0)



/* init parse structure */
void parse_init(struct parse *parse, void *user_data,
        parse_tag_hook_t *tag_hook, parse_data_hook_t *data_hook) {
    parse->state = STATE_DATA;

    parse->buf = NULL;
    parse->buf_size = 0;
    parse->buf_pos = 0;
    parse->name_start = -1;
    parse->name_length = 0;

    parse->line = 1;
    parse->col = 1;

    parse->work_dir[0] = '\0';
    parse->user_data = user_data;
    parse->tag_hook = tag_hook;
    parse->data_hook = data_hook;
}


int parse_make_work_dir(struct parse *parse, const char *filename) {
    char path_buf[PATH_MAX + 1], *path;

    if(strlen(filename) >= sizeof(path_buf)) {
        errno = ENAMETOOLONG;
        return -1;
    }
    strcpy(path_buf, filename);
    path = dirname(path_buf);
    strcpy(parse->work_dir, path);

    return 0;
}


/* release parse structure */
int parse_release(struct parse *parse) {
    int err = 0;

    /* flush TAG buffer */
    if(BUF_IS_NOT_EMPTY(parse)) {
        if(parse->data_hook != NULL) {
            err = parse->data_hook(parse, parse->buf, parse->buf_pos);
        }
    }

    /* free TAG buffer */
    free(parse->buf);

    return err;
}


/* parse template chunk */
int parse_routine(struct parse *parse, const char *data, int length) {
    char c;
    off_t n, r, off = 0;
    int ret;


    if(length < 0) {
        length = strlen(data);
    }

    r = 1;
    n = 0;
    while(n < length) {
        c = data[n];
        switch(parse->state) {
        case STATE_DATA:
            if(c == '<') {
                if(BUF_IS_NOT_EMPTY(parse)) {
                    /* flush data */
                    if(parse->data_hook != NULL) {
                        if((ret = parse->data_hook(parse, parse->buf, parse->buf_pos)) < 0) {
                            return ret;
                        }
                    }
                    BUF_RESET(parse);
                }
                if(off < n) {
                    if((ret = parse->data_hook(parse, data + off, n - off)) < 0) {
                        return ret;
                    }
                }
                off = n;
                parse->tag_line = parse->line;
                parse->tag_col = parse->col;
                parse->state = STATE_TMPL_0;
            }
            break;
        case STATE_TMPL_CLOSE:
            if(c == 'T') {
                parse->tag_type = TAG_CLOSE;
                parse->state = STATE_TMPL_1;
            }
            else
                parse->state = STATE_DATA;
            break;
        case STATE_TMPL_0:
            if(c == 'T') {
                parse->tag_type = TAG_OPEN;
                parse->state = STATE_TMPL_1;
            }
            else if(c == '/')
                parse->state = STATE_TMPL_CLOSE;
            else
                parse->state = STATE_DATA;
            break;
        case STATE_TMPL_1:
            if(c == 'M')
                parse->state = STATE_TMPL_2;
            else
                parse->state = STATE_DATA;
            break;
        case STATE_TMPL_2:
            if(c == 'P')
                parse->state = STATE_TMPL_3;
            else
                parse->state = STATE_DATA;
            break;
        case STATE_TMPL_3:
            if(c == 'L')
                parse->state = STATE_TMPL_4;
            else
                parse->state = STATE_DATA;
            break;
        case STATE_TMPL_4:
            if(c == '_')
                parse->state = STATE_TMPL_5;
            else
                parse->state = STATE_DATA;
            break;
        case STATE_TMPL_5:
            if(c == 'V')
                parse->state = STATE_VAR_0;
            else if(c == 'L')
                parse->state = STATE_LOOP_0;
            else if(c == 'I')
                parse->state = STATE_INCLUDE_IF;
            else if(c == 'U')
                parse->state = STATE_UNLESS_0;
            else if(c == 'E')
                parse->state = STATE_ELSE_0;
            else
                parse->state = STATE_DATA;
            break;

        case STATE_VAR_0:
            if(c == 'A')
                parse->state = STATE_VAR_1;
            else
                parse->state = STATE_DATA;
            break;
        case STATE_VAR_1:
            if(c == 'R') {
                if(parse->tag_type == TAG_OPEN) {
                    parse->tag = TAG_VAR;
                    parse->state = STATE_NAME_SPACE;
                    break;
                }
            }
            parse->state = STATE_DATA;
            break;

        case STATE_LOOP_0:
            if(c == 'O')
                parse->state = STATE_LOOP_1;
            else
                parse->state = STATE_DATA;
            break;
        case STATE_LOOP_1:
            if(c == 'O')
                parse->state = STATE_LOOP_2;
            else
                parse->state = STATE_DATA;
            break;
        case STATE_LOOP_2:
            if(c == 'P') {
                if(parse->tag_type == TAG_OPEN) {
                    parse->tag = TAG_LOOP;
                    parse->state = STATE_NAME_SPACE;
                }
                else {
                    parse->tag = TAG_LOOP_CLOSE;
                    parse->state = STATE_TAG_END;
                }
            }
            else
                parse->state = STATE_DATA;
            break;

        case STATE_INCLUDE_IF:
            if(c == 'N')
                parse->state = STATE_INCLUDE_0;
            else if(c == 'F') {
                if(parse->tag_type == TAG_OPEN) {
                    parse->tag = TAG_IF;
                    parse->state = STATE_NAME_SPACE;
                }
                else {
                    parse->tag = TAG_IF_CLOSE;
                    parse->state = STATE_TAG_END;
                }
            }
            else
                parse->state = STATE_DATA;
            break;

        case STATE_INCLUDE_0:
            if(c == 'C')
                parse->state = STATE_INCLUDE_1;
            else
                parse->state = STATE_DATA;
            break;
        case STATE_INCLUDE_1:
            if(c == 'L')
                parse->state = STATE_INCLUDE_2;
            else
                parse->state = STATE_DATA;
            break;
        case STATE_INCLUDE_2:
            if(c == 'U')
                parse->state = STATE_INCLUDE_3;
            else
                parse->state = STATE_DATA;
            break;
        case STATE_INCLUDE_3:
            if(c == 'D')
                parse->state = STATE_INCLUDE_4;
            else
                parse->state = STATE_DATA;
            break;
        case STATE_INCLUDE_4:
            if(c == 'E') {
                if(parse->tag_type == TAG_OPEN) {
                    parse->tag = TAG_INCLUDE;
                    parse->state = STATE_NAME_SPACE;
                    break;
                }
            }
            parse->state = STATE_DATA;
            break;

        case STATE_ELSE_0:
            if(c == 'L')
                parse->state = STATE_ELSE_1;
            else
                parse->state = STATE_DATA;
            break;
        case STATE_ELSE_1:
            if(c == 'S')
                parse->state = STATE_ELSE_2;
            else
                parse->state = STATE_DATA;
            break;
        case STATE_ELSE_2:
            if(c == 'E') {
                if(parse->tag_type == TAG_OPEN) {
                    parse->tag = TAG_ELSE;
                    parse->state = STATE_TAG_END;
                    break;
                }
            }
            parse->state = STATE_DATA;
            break;

        case STATE_UNLESS_0:
            if(c == 'N')
                parse->state = STATE_UNLESS_1;
            else
                parse->state = STATE_DATA;
            break;
        case STATE_UNLESS_1:
            if(c == 'L')
                parse->state = STATE_UNLESS_2;
            else
                parse->state = STATE_DATA;
            break;
        case STATE_UNLESS_2:
            if(c == 'E')
                parse->state = STATE_UNLESS_3;
            else
                parse->state = STATE_DATA;
            break;
        case STATE_UNLESS_3:
            if(c == 'S')
                parse->state = STATE_UNLESS_4;
            else
                parse->state = STATE_DATA;
            break;
        case STATE_UNLESS_4:
            if(c == 'S') {
                if(parse->tag_type == TAG_OPEN) {
                    parse->tag = TAG_UNLESS;
                    parse->state = STATE_NAME_SPACE;
                }
                else {
                    parse->tag = TAG_UNLESS_CLOSE;
                    parse->state = STATE_TAG_END;
                }
            }
            else
                parse->state = STATE_DATA;
            break;

        case STATE_NAME_SPACE:
            if(NOT_WHITESPACE(c)) {
                if(c == 'N')
                    parse->state = STATE_NAME_0;
                else
                    parse->state = STATE_DATA;
            }
            break;
        case STATE_NAME_0:
            if(c == 'A')
                parse->state = STATE_NAME_1;
            else
                parse->state = STATE_DATA;
            break;
        case STATE_NAME_1:
            if(c == 'M')
                parse->state = STATE_NAME_2;
            else
                parse->state = STATE_DATA;
            break;
        case STATE_NAME_2:
            if(c == 'E')
                parse->state = STATE_NAME_3;
            else
                parse->state = STATE_DATA;
            break;
        case STATE_NAME_3:
            if(c == '=')
                parse->state = STATE_NAME_VAL;
            else
                parse->state = STATE_DATA;
            break;
        case STATE_NAME_VAL:
            parse->name_start = parse->buf_pos + (n - off);
            if(c == '"') {
                parse->name_start++;
                parse->state = STATE_NAME_VAL_QUOT;
            }
            else {
                parse->state = STATE_NAME_VAL_NOQUOT;
                r = 0;
            }
            break;
        case STATE_NAME_VAL_QUOT:
            if(c == '"') {
                parse->name_length = (parse->buf_pos + (n - off)) - parse->name_start;
                parse->state = STATE_TAG_END;
            }
            break;
        case STATE_NAME_VAL_NOQUOT:
            if(IS_WHITESPACE(c) || c == '>') {
                parse->name_length = (parse->buf_pos + (n - off)) - parse->name_start;
                parse->state = STATE_TAG_END;
                r = 0;
            }
            break;

        case STATE_TAG_END:
            if(c == '>') {
                /* add TAG remainder to buffer */
                BUFF_APPEND(parse, (char *)(data + off), n - off);

                if(parse->tag_hook != NULL) {
                    char *var;
                    if(parse->name_start >= 0) {
                        var = parse->buf + parse->name_start;
                        parse->buf[parse->name_start + parse->name_length] = '\0';
                    }
                    else
                        var = NULL;
                    ret = parse->tag_hook(parse, parse->tag, var);
                    if(ret < 0)
                        return ret;
                }
                BUF_RESET(parse);

                off = n + 1;

                parse->state = STATE_DATA;
            }
            else if(NOT_WHITESPACE(c))
                parse->state = STATE_DATA;
            break;
        }

        if(r) {
            n++;

            /* calcilate statistics */
            if(c == '\n') {
                parse->line++;
                parse->col = 1;
            }
            else
                parse->col++;
        }
        else
            r = 1;
    }

    if(parse->state != STATE_DATA) {
        /* push TAG to buffer */
        BUFF_APPEND(parse, data + off, length - off);
    }
    else {
        /* flush TAG buffer */
        if(BUF_IS_NOT_EMPTY(parse)) {
            if(parse->data_hook != NULL) {
                if((ret = parse->data_hook(parse, parse->buf, parse->buf_pos)) < 0)
                    return ret;
            }
            BUF_RESET(parse);
        }

        /* fluch data */
        if(parse->data_hook != NULL && off < length) {
            if((ret = parse->data_hook(parse, data + off, length - off)) < 0)
                return ret;
        }
    }

    return 0;
}


/* get current TAG position */
inline int parse_tag_col(struct parse *parse) {
    return parse->tag_col;
}


inline int parse_tag_row(struct parse *parse) {
    return parse->tag_line;
}
